from typing import Generic, TypeVar

from rainbowadn.core import *
from rainbowadn.nullability import *
from .block import *
from .blockchainprotocol import BlockChainProtocol
from .chaincollectionfactory import ChainCollectionFactory
from .chaincollectioninterface import ChainCollectionInterface

__all__ = ('BlockChain', 'BlockChainFactory',)

HeaderType = TypeVar('HeaderType')
StateType = TypeVar('StateType')
ActualStateType = TypeVar('ActualStateType')


class BlockChain(
    ChainCollectionInterface[
        Block[
            HeaderType,
            StateType
        ],
        HeaderType,
        ActualStateType
    ],
    Generic[HeaderType, ActualStateType, StateType],
):
    def __init__(
            self,
            reference: NullableReference[
                Block[
                    HeaderType,
                    StateType
                ]
            ],
            protocol: BlockChainProtocol[HeaderType, StateType, ActualStateType],
    ):
        assert isinstance(reference, NullableReference)
        assert isinstance(protocol, BlockChainProtocol)
        super().__init__(reference)
        self.protocol = protocol

    def factory(self) -> ChainCollectionFactory[
        Block[
            HeaderType,
            StateType
        ],
        HeaderType,
        ActualStateType
    ]:
        return BlockChainFactory(self.protocol)

    async def add(
            self,
            header: HashPoint[HeaderType]
    ) -> ChainCollectionInterface[
        Block[
            HeaderType,
            StateType
        ],
        HeaderType,
        ActualStateType
    ]:
        assert isinstance(header, HashPoint)
        return self.factory().from_reference(
            NullableReference.off(
                await self._add(header)
            )
        )

    async def state(
            self
    ) -> NullableReference[
        StateType
    ]:
        if self.reference.null():
            return NullableReference(
                Null(),
                self.protocol.state_factory
            )
        else:
            block: Block[HeaderType, StateType] = await self.reference.resolve()
            assert isinstance(block, Block)
            return NullableReference.of(block.state)

    async def _add(
            self,
            header: HashPoint[HeaderType]
    ) -> Block[
        HeaderType,
        StateType
    ]:
        assert isinstance(header, HashPoint)
        return Block(
            self.reference,
            header,
            await self.protocol.state_protocol.derive(
                await self.state(),
                header,
            )
        )

    async def previous(
            self
    ) -> ChainCollectionInterface[
        Block[
            HeaderType,
            StateType
        ],
        HeaderType,
        ActualStateType
    ]:
        assert not self.reference.null()
        block: Block[
            HeaderType,
            StateType
        ] = await self.reference.resolve()
        assert isinstance(block, Block)
        return self.factory().from_reference(block.previous)

    async def verify(self) -> bool:
        if self.reference.null():
            return True
        else:
            previous: ChainCollectionInterface[
                Block[HeaderType, StateType], HeaderType, ActualStateType
            ] = await self.previous()
            assert isinstance(previous, ChainCollectionInterface)
            assert_trues(
                await gather(
                    self.verify_link(previous.reference),
                    previous.verify(),
                )
            )
            return True

    async def _verify_link(
            self,
            block: Block[
                HeaderType,
                StateType
            ],
            previous: NullableReference[
                Block[
                    HeaderType,
                    StateType
                ]
            ]
    ) -> bool:
        assert isinstance(block, Block)
        assert isinstance(previous, NullableReference)
        assert_eq(block.previous, previous)
        previous_chain = self.factory().from_reference(previous)
        assert isinstance(previous_chain, BlockChain)
        assert_true(
            await self.protocol.state_protocol.verify(
                await previous_chain.state(),
                block.header,
                block.state,
            )
        )
        return True

    def _actual_state_factory(self) -> RainbowFactory[ActualStateType]:
        return self.protocol.actual_state_factory()

    async def _actual_state(
            self,
            block: Block[
                HeaderType,
                StateType
            ]
    ) -> HashPoint[ActualStateType]:
        assert isinstance(block, Block)
        return await self.protocol.actual_state(await block.state.resolve())

    def loose(self) -> ChainCollectionInterface[
        Block[
            HeaderType,
            StateType
        ],
        HeaderType,
        ActualStateType
    ]:
        return self


class BlockChainFactory(
    ChainCollectionFactory[
        Block[
            HeaderType,
            StateType
        ],
        HeaderType,
        ActualStateType
    ],
    Generic[HeaderType, ActualStateType, StateType]
):
    def __init__(
            self,
            protocol: BlockChainProtocol[HeaderType, StateType, ActualStateType],
    ):
        assert isinstance(protocol, BlockChainProtocol)
        self.protocol = protocol

    def empty(self) -> BlockChain[HeaderType, ActualStateType, StateType]:
        return BlockChain(
            NullableReference(
                Null(),
                BlockFactory(
                    self.protocol.header_factory,
                    self.protocol.state_factory
                )
            ),
            self.protocol
        )

    def from_reference(
            self,
            reference: NullableReference[
                Block[
                    HeaderType,
                    StateType
                ],
            ]
    ) -> BlockChain[
        HeaderType, ActualStateType, StateType
    ]:
        assert isinstance(reference, NullableReference)
        return BlockChain(
            reference,
            self.protocol
        )
