from typing import Generic, Iterable, TypeVar

from rainbowadn.core import *
from rainbowadn.nullability import *
from .stageprotocol import StageProtocol

__all__ = (
    'StageStage',
    'StageStageFactory',
    'StateStage',
    'StateStageFactory',
)

HeaderType = TypeVar('HeaderType')
BaseStateType = TypeVar('BaseStateType')
StageType = TypeVar('StageType')


class Stage(Generic[HeaderType, BaseStateType, StageType]):
    def __init__(self, protocol: StageProtocol[HeaderType, BaseStateType, StageType]):
        assert isinstance(protocol, StageProtocol)
        self.protocol = protocol


class StageStage(
    RecursiveMentionable,
    Stage[HeaderType, BaseStateType, StageType],
    Generic[HeaderType, BaseStateType, StageType],
):
    def __init__(
            self,
            protocol: StageProtocol[HeaderType, BaseStateType, StageType],
            previous: NullableReference['StageStage[HeaderType, BaseStateType, StageType]'],
            stage: HashPoint[StageType],
    ):
        assert isinstance(protocol, StageProtocol)
        assert isinstance(previous, NullableReference)
        assert isinstance(stage, HashPoint)
        super().__init__(protocol)
        self.previous = previous
        self.stage = stage

    @classmethod
    async def _previous_state(
            cls,
            previous: NullableReference['StateStage[HeaderType, BaseStateType, StageType]'],
            base_factory: RainbowFactory[BaseStateType],
    ) -> NullableReference[BaseStateType]:
        assert isinstance(previous, NullableReference)
        assert isinstance(base_factory, RainbowFactory)
        if previous.null():
            return NullableReference(Null(), base_factory)
        else:
            state_stage: StateStage[HeaderType, BaseStateType, StageType] = await previous.resolve()
            assert isinstance(state_stage, StateStage)
            return NullableReference.of(state_stage.state)

    async def verify(
            self,
            previous: NullableReference['StateStage[HeaderType, BaseStateType, StageType]'],
            header: HashPoint[HeaderType],
            base_factory: RainbowFactory[BaseStateType],
    ) -> bool:
        assert isinstance(previous, NullableReference)
        assert isinstance(header, HashPoint)
        assert isinstance(base_factory, RainbowFactory)
        if self.previous.null():
            return await self.protocol.verify_header(
                await self._previous_state(previous, base_factory),
                header,
                self.stage
            )
        else:
            previous_stage: StageStage[HeaderType, BaseStateType, StageType] = await self.previous.resolve()
            assert isinstance(previous_stage, StageStage)
            assert_trues(
                await gather(
                    self.protocol.verify_stage(
                        previous_stage.stage,
                        self.stage
                    ),
                    previous_stage.verify(
                        previous,
                        header,
                        base_factory
                    ),
                )
            )
            return True

    def points(self) -> Iterable[HashPoint]:
        return [*self.previous.points(), self.stage]

    def __bytes__(self):
        return bytes(self.previous) + bytes(self.stage)

    def __factory__(self) -> RainbowFactory['StageStage[HeaderType, BaseStateType, StageType]']:
        return StageStageFactory(
            self.protocol,
            self.stage.factory
        )

    async def str(self, tab: int) -> str:
        assert isinstance(tab, int)
        previous_str, stage_str = await gather(
            self.previous.str(tab),
            hash_point_format(self.stage, tab)
        )
        assert isinstance(previous_str, str)
        assert isinstance(stage_str, str)
        return f'{previous_str}' \
               f'{tabulate(tab)}{stage_str}'


class StageStageFactory(RainbowFactory[StageStage[HeaderType, BaseStateType, StageType]]):
    def __init__(
            self,
            protocol: StageProtocol[HeaderType, BaseStateType, StageType],
            stage_factory: RainbowFactory[StageType]
    ):
        assert isinstance(protocol, StageProtocol)
        assert isinstance(stage_factory, RainbowFactory)
        self.protocol = protocol
        self.stage_factory = stage_factory

    def from_bytes(self, source: bytes, resolver: HashResolver) -> StageStage[HeaderType, BaseStateType, StageType]:
        assert isinstance(source, bytes)
        assert isinstance(resolver, HashResolver)
        return StageStage(
            self.protocol,
            NullableReferenceFactory(self).from_bytes(source[:HashPoint.HASH_LENGTH], resolver),
            ResolverOrigin(self.stage_factory, source[HashPoint.HASH_LENGTH:], resolver).hash_point()
        )


class StateStage(
    RecursiveMentionable,
    Stage[HeaderType, BaseStateType, StageType],
    Generic[HeaderType, BaseStateType, StageType]
):
    def __init__(
            self,
            protocol: StageProtocol[HeaderType, BaseStateType, StageType],
            previous: HashPoint[StageStage[HeaderType, BaseStateType, StageType]],
            state: HashPoint[BaseStateType],
            stage_factory: RainbowFactory[StageType]
    ):
        assert isinstance(protocol, StageProtocol)
        assert isinstance(previous, HashPoint)
        assert isinstance(state, HashPoint)
        assert isinstance(stage_factory, RainbowFactory)
        super().__init__(protocol)
        self.previous = previous
        self.state = state
        self.stage_factory = stage_factory

    async def verify(
            self,
            previous: NullableReference['StateStage[HeaderType, BaseStateType, StageType]'],
            header: HashPoint[HeaderType]
    ) -> bool:
        assert isinstance(previous, NullableReference)
        assert isinstance(header, HashPoint)
        previous_stage: StageStage[HeaderType, BaseStateType, StageType] = await self.previous.resolve()
        assert isinstance(previous_stage, StageStage)
        assert_trues(
            await gather(
                self.protocol.verify_state(
                    previous_stage.stage,
                    self.state
                ),
                previous_stage.verify(
                    previous,
                    header,
                    self.state.factory
                )
            )
        )
        return True

    def points(self) -> Iterable[HashPoint]:
        return [self.previous, self.state]

    def __bytes__(self):
        return bytes(self.previous) + bytes(self.state)

    def __factory__(self) -> RainbowFactory['StateStage[HeaderType, BaseStateType, StageType]']:
        return StateStageFactory(
            self.protocol,
            self.stage_factory,
            self.state.factory
        )

    async def str(self, tab: int) -> str:
        assert isinstance(tab, int)
        previous_str, state_str = await gather(
            hash_point_format(self.previous, tab),
            hash_point_format(self.state, tab),
        )
        assert isinstance(previous_str, str)
        assert isinstance(state_str, str)
        return f'{previous_str}' \
               f'{tabulate(tab)}{state_str}'


class StateStageFactory(RainbowFactory[StateStage[HeaderType, BaseStateType, StageType]]):
    def __init__(
            self,
            protocol: StageProtocol[HeaderType, BaseStateType, StageType],
            stage_factory: RainbowFactory[StageType],
            state_factory: RainbowFactory[BaseStateType]
    ):
        assert isinstance(protocol, StageProtocol)
        assert isinstance(stage_factory, RainbowFactory)
        assert isinstance(state_factory, RainbowFactory)
        self.protocol = protocol
        self.stage_factory = stage_factory
        self.state_factory = state_factory

    def from_bytes(self, source: bytes, resolver: HashResolver) -> StateStage[HeaderType, BaseStateType, StageType]:
        assert isinstance(source, bytes)
        assert isinstance(resolver, HashResolver)
        return StateStage(
            self.protocol,
            ResolverOrigin(
                StageStageFactory(
                    self.protocol,
                    self.stage_factory
                ),
                source[:HashPoint.HASH_LENGTH],
                resolver
            ).hash_point(),
            ResolverOrigin(self.state_factory, source[HashPoint.HASH_LENGTH:], resolver).hash_point(),
            self.stage_factory
        )
