from typing import Generic, Iterable, TypeVar

from rainbowadn.core import *

__all__ = ('Pair', 'PairFactory',)

E0Type = TypeVar('E0Type')
E1Type = TypeVar('E1Type')


class Pair(
    RecursiveMentionable,
    Generic[E0Type, E1Type]
):
    def __init__(self, element0: HashPoint[E0Type], element1: HashPoint[E1Type]):
        assert isinstance(element0, HashPoint)
        assert isinstance(element1, HashPoint)
        self.element0 = element0
        self.element1 = element1

    def points(self) -> Iterable[HashPoint]:
        return [self.element0, self.element1]

    def __bytes__(self):
        return bytes(self.element0) + bytes(self.element1)

    def __factory__(self) -> RainbowFactory['Pair[E0Type, E1Type]']:
        return PairFactory(self.element0.factory, self.element1.factory)

    async def str(self, tab: int) -> str:
        assert isinstance(tab, int)
        e0_str, e1_str = await gather(
            hash_point_format(self.element0, tab),
            hash_point_format(self.element1, tab),
        )
        assert isinstance(e0_str, str)
        assert isinstance(e1_str, str)
        return f'(pair)' \
               f'{tabulate(tab)}{e0_str}' \
               f'{tabulate(tab)}{e1_str}'


class PairFactory(
    RainbowFactory[Pair[E0Type, E1Type]],
    Generic[E0Type, E1Type]
):
    def __init__(
            self,
            e0_factory: RainbowFactory[E0Type],
            e1_factory: RainbowFactory[E1Type],
    ):
        assert isinstance(e0_factory, RainbowFactory)
        assert isinstance(e1_factory, RainbowFactory)
        self.e0_factory = e0_factory
        self.e1_factory = e1_factory

    def from_bytes(self, source: bytes, resolver: HashResolver) -> Pair[E0Type, E1Type]:
        assert isinstance(source, bytes)
        assert isinstance(resolver, HashResolver)
        return Pair(
            ResolverOrigin(self.e0_factory, source[:HashPoint.HASH_LENGTH], resolver).hash_point(),
            ResolverOrigin(self.e1_factory, source[HashPoint.HASH_LENGTH:], resolver).hash_point(),
        )

    def loose(self) -> RainbowFactory[Pair[E0Type, E1Type]]:
        return self
