from typing import Generic, Iterable, TypeVar

from nacl.bindings import crypto_hash_sha256
from nacl.secret import SecretBox

from rainbowadn.core import *

__all__ = ('Encrypted', 'EncryptedFactory')

EncryptedType = TypeVar('EncryptedType')


class Encrypted(RecursiveMentionable, Generic[EncryptedType]):
    """unstable"""

    def __init__(
            self,
            key: bytes,
            resolution: tuple[HashPoint['Encrypted'], ...],
            mapping: dict[bytes, HashPoint['Encrypted']],
            decrypted: EncryptedType
    ):
        assert isinstance(key, bytes)
        assert isinstance(resolution, tuple)
        assert isinstance(mapping, dict)
        assert isinstance(decrypted, Mentionable)

        self.factory: RainbowFactory[EncryptedType] = decrypted.__factory__()
        assert isinstance(self.factory, RainbowFactory)
        self.key = key

        self.resolution = resolution
        self.decrypted: EncryptedType = decrypted

        self.mapping = mapping

    def points(self) -> Iterable[HashPoint]:
        return self.resolution

    @classmethod
    async def encrypt(cls, decrypted: EncryptedType, key: bytes) -> 'Encrypted[EncryptedType]':
        assert isinstance(decrypted, Mentionable)
        assert isinstance(key, bytes)
        hashpoints = tuple(decrypted.points()) if isinstance(decrypted, RecursiveMentionable) else ()
        resolution = tuple(
            await gather(
                *(
                    cls.encrypt_hashpoint(hash_point, key) for hash_point in hashpoints
                )
            )
        )
        return cls.construct(
            key,
            resolution,
            hashpoints,
            decrypted
        )

    @classmethod
    async def encrypt_hashpoint(
            cls, hash_point: HashPoint[EncryptedType], key: bytes
    ) -> HashPoint['Encrypted[EncryptedType]']:
        assert isinstance(hash_point, HashPoint)
        assert isinstance(key, bytes)
        if isinstance(hash_point.origin, ResolverOrigin):
            resolver: HashResolver = hash_point.origin.resolver
            assert isinstance(resolver, HashResolver)
            if isinstance(resolver, EncryptedResolver) and resolver.key == key:
                return ShortcutOrigin(
                    hash_point.factory,
                    resolver.mapping[hash_point.point],
                    key
                ).hash_point()
        return HashPoint.of(await cls.encrypt(await hash_point.resolve(), key))

    @classmethod
    def construct(
            cls,
            key: bytes,
            resolution: tuple[HashPoint['Encrypted'], ...],
            hashpoints: tuple[HashPoint, ...],
            decrypted: EncryptedType
    ) -> 'Encrypted[EncryptedType]':
        assert isinstance(key, bytes)
        assert isinstance(resolution, tuple)
        assert isinstance(hashpoints, tuple)
        assert isinstance(decrypted, Mentionable)
        mapping: dict[bytes, HashPoint[Encrypted]] = {
            hash_point.point: encrypted for hash_point, encrypted in zip(hashpoints, resolution)
        }
        return Encrypted(
            key,
            resolution,
            mapping,
            decrypted.__factory__().from_bytes(
                bytes(decrypted),
                EncryptedResolver(mapping, key)
            )
        )

    def __bytes__(self):
        source: bytes = len(self.resolution).to_bytes(8, 'little') + b''.join(
            encrypted.point
            for
            encrypted
            in self.resolution
        ) + bytes(self.decrypted)
        nonce: bytes = crypto_hash_sha256(self.key + source)[:24]
        return SecretBox(self.key).encrypt(source, nonce=nonce)

    def __factory__(self) -> RainbowFactory['Encrypted[EncryptedType]']:
        return EncryptedFactory(self.factory, self.key)


class EncryptedFactory(RainbowFactory[Encrypted[EncryptedType]], Generic[EncryptedType]):
    def __init__(self, factory: RainbowFactory[EncryptedType], key: bytes):
        assert isinstance(factory, RainbowFactory)
        assert isinstance(key, bytes)
        self.factory = factory
        self.key = key

    def from_bytes(self, source: bytes, resolver: HashResolver) -> Encrypted[EncryptedType]:
        assert isinstance(source, bytes)
        assert isinstance(resolver, HashResolver)
        plain: bytes = SecretBox(self.key).decrypt(source)
        resolution_size: int = int.from_bytes(plain[:8], 'little')
        decrypted: EncryptedType = self.factory.from_bytes(
            plain[8 + resolution_size * HashPoint.HASH_LENGTH:],
            resolver,
        )
        assert isinstance(decrypted, Mentionable)
        hashpoints = tuple(decrypted.points()) if isinstance(decrypted, RecursiveMentionable) else ()
        assert_eq(len(hashpoints), resolution_size)
        resolution: tuple[HashPoint[Encrypted], ...] = tuple(
            ResolverOrigin(
                EncryptedFactory(
                    hash_point.factory,
                    self.key
                ),
                plain[8 + i * HashPoint.HASH_LENGTH: 8 + (i + 1) * HashPoint.HASH_LENGTH],
                resolver
            ).hash_point() for i, hash_point in enumerate(hashpoints)
        )
        return Encrypted.construct(
            self.key,
            resolution,
            hashpoints,
            decrypted
        )


class EncryptedResolver(HashResolver):
    def __init__(self, mapping: dict[bytes, HashPoint[Encrypted]], key: bytes):
        assert isinstance(mapping, dict)
        assert isinstance(key, bytes)
        self.mapping = mapping
        self.key = key

    async def resolve(self, point: bytes) -> tuple[bytes, 'HashResolver']:
        assert isinstance(point, bytes)
        encrypted: Encrypted = await self.mapping[point].resolve()
        assert isinstance(encrypted, Encrypted)
        return HashPoint.bytes_of_mentioned(encrypted.decrypted), EncryptedResolver(encrypted.mapping, self.key)


class ShortcutOrigin(Origin[Encrypted[EncryptedType]], Generic[EncryptedType]):
    def __init__(self, factory: RainbowFactory[EncryptedType], hash_point: HashPoint[Encrypted], key: bytes):
        assert isinstance(factory, RainbowFactory)
        assert isinstance(hash_point, HashPoint)
        assert isinstance(key, bytes)
        self.factory: RainbowFactory[Encrypted[EncryptedType]] = EncryptedFactory(factory, key)
        assert isinstance(self.factory, RainbowFactory)
        self.hashpoint = hash_point
        super().__init__(self.factory)

    async def resolve(self) -> Encrypted[EncryptedType]:
        source: Encrypted = await self.hashpoint.resolve()
        assert isinstance(source, Encrypted)
        encrypted: Encrypted[EncryptedType] = self.factory.from_bytes(bytes(source), ShortcutResolver(source))
        assert isinstance(encrypted, Encrypted)
        assert_eq(HashPoint.of(encrypted), self.hashpoint)
        return encrypted

    def hash_point(self) -> HashPoint[Encrypted[EncryptedType]]:
        return HashPoint(
            self.hashpoint.point,
            self
        )


class ShortcutResolver(HashResolver):
    def __init__(self, encrypted: Encrypted):
        assert isinstance(encrypted, Encrypted)
        self.mapping: dict[bytes, HashPoint[Encrypted]] = {
            hash_point.point: hash_point for hash_point in encrypted.resolution
        }

    async def resolve(self, point: bytes) -> tuple[bytes, 'HashResolver']:
        assert isinstance(point, bytes)
        resolved: Encrypted = await self.mapping[point].resolve()
        assert isinstance(resolved, Encrypted)
        return (
            HashPoint.bytes_of_mentioned(resolved),
            ShortcutResolver(resolved)
        )
