from typing import Generic, Iterable, TypeVar

from rainbowadn.core import *
from .notnull import NotNull
from .null import Null
from .nullable import Nullable

__all__ = ('NullableReference', 'NullableReferenceFactory',)

Referenced = TypeVar('Referenced')


class NullableReference(RecursiveMentionable, Generic[Referenced]):
    def __factory__(self) -> RainbowFactory['NullableReference[Referenced]']:
        return NullableReferenceFactory(self.factory)

    def points(self) -> Iterable[HashPoint]:
        if self.null():
            return []
        else:
            return [self.hashpoint()]

    def __bytes__(self):
        if self.null():
            return HashPoint.NULL_HASH
        else:
            return self.hashpoint().point

    def __init__(
            self, reference: Nullable[HashPoint[Referenced]], factory: RainbowFactory[Referenced]
    ):
        assert isinstance(reference, Nullable)
        assert isinstance(factory, RainbowFactory)
        self.reference = reference
        self.factory = factory

    @classmethod
    def of(cls, hash_point: HashPoint[Referenced]) -> 'NullableReference[Referenced]':
        assert isinstance(hash_point, HashPoint)
        return cls(NotNull(hash_point), hash_point.factory)

    @classmethod
    def off(cls, value: Referenced) -> 'NullableReference[Referenced]':
        assert isinstance(value, Mentionable)
        return cls.of(HashPoint.of(value))

    async def str(self, tab: int) -> str:
        assert isinstance(tab, int)
        if self.null():
            return f'-'
        else:
            return f'{await hash_point_format(self.hashpoint(), tab)}'

    def __eq__(self, other):
        if isinstance(other, NullableReference):
            return self.reference == other.reference
        else:
            return NotImplemented

    def null(self) -> bool:
        return self.reference.null()

    def hashpoint(self) -> HashPoint[Referenced]:
        assert not self.null()
        return self.reference.resolve()

    async def resolve(self) -> Referenced:
        assert not self.null()
        return await self.hashpoint().resolve()


class NullableReferenceFactory(RainbowFactory[NullableReference[Referenced]], Generic[Referenced]):
    def __init__(self, factory: RainbowFactory[Referenced]):
        assert isinstance(factory, RainbowFactory)
        self.factory = factory

    def from_bytes(self, source: bytes, resolver: HashResolver) -> NullableReference[Referenced]:
        assert isinstance(source, bytes)
        assert isinstance(resolver, HashResolver)
        if source == HashPoint.NULL_HASH:
            return NullableReference(Null(), self.factory)
        else:
            return NullableReference.of(
                ResolverOrigin(self.factory, source, resolver).hash_point()
            )

    def loose(self) -> RainbowFactory[NullableReference[Referenced]]:
        return self
