import functools
from contextlib import ExitStack
from typing import Callable, Generic, TypeVar

__all__ = ('Instrumentation',)

IType = TypeVar('IType')


class Instrumentation(Generic[IType]):
    deinstrumentation = {}
    method: Callable
    wrap: Callable

    def __init__(self, target, methodname: str):
        assert isinstance(methodname, str)
        assert callable(getattr(target, methodname))
        self.target = target
        self.methodname = methodname

    def instrument(self, method, *args, **kwargs):
        raise NotImplementedError

    def __enter__(self: IType) -> IType:
        assert not hasattr(self, 'method')
        self.method = getattr(self.target, self.methodname)
        assert callable(self.method)

        @functools.wraps(self.method)
        def wrap(*args, **kwargs):
            return self.instrument(self.method, *args, **kwargs)

        self.wrap = wrap

        setattr(self.target, self.methodname, self.wrap)

        return self

    def schedule_deinstrumentation(self):
        self.deinstrumentation[self.wrap] = self.method
        del self.wrap
        del self.method

    def deinstrument(self):
        while (method := getattr(self.target, self.methodname)) in self.deinstrumentation:
            setattr(self.target, self.methodname, self.deinstrumentation.pop(method))

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.schedule_deinstrumentation()
        self.deinstrument()

    def enter(self: IType, stack: ExitStack) -> IType:
        return stack.enter_context(self)
