from typing import AsyncIterable, Iterable

from rainbowadn.atomic import *
from rainbowadn.collection.comparison import *
from rainbowadn.collection.keymetadata import *
from rainbowadn.collection.trees.binary import *
from rainbowadn.core import *
from rainbowadn.nullability import *
from .algo import MINT_CONST
from .subject import Subject
from .transaction import *

__all__ = ('BankState',)


class BankState(RecursiveMentionable, StaticMentionable):
    def __init__(
            self,
            minted: NullableReference[BinaryTree[KeyMetadata[Coin, Integer]]],
            used: NullableReference[BinaryTree[KeyMetadata[Coin, Integer]]],
            miner: NullableReference[Subject],
            length: HashPoint[Integer]
    ):
        assert isinstance(miner, NullableReference)
        assert isinstance(used, NullableReference)
        assert isinstance(miner, NullableReference)
        assert isinstance(length, HashPoint)
        self.minted = minted
        self.used = used
        self.miner = miner
        self.length = length

    def points(self) -> Iterable[HashPoint]:
        return [*self.minted.points(), *self.used.points(), *self.miner.points(), self.length]

    def __bytes__(self):
        return bytes(self.minted) + bytes(self.used) + bytes(self.miner) + bytes(self.length)

    @classmethod
    def from_bytes(cls, source: bytes, resolver: HashResolver) -> 'BankState':
        assert isinstance(source, bytes)
        assert isinstance(resolver, HashResolver)
        reference_factory: RainbowFactory[
            NullableReference[BinaryTree[KeyMetadata[Coin, Integer]]]
        ] = NullableReferenceFactory(
            BinaryTreeFactory(
                KeyMetadataFactory(
                    StaticFactory(Coin),
                    StaticFactory(Integer),
                ).loose()
            ).loose()
        ).loose()
        assert isinstance(reference_factory, RainbowFactory)
        return cls(
            reference_factory.from_bytes(source[:HashPoint.HASH_LENGTH], resolver),
            reference_factory.from_bytes(source[HashPoint.HASH_LENGTH:2 * HashPoint.HASH_LENGTH], resolver),
            NullableReferenceFactory(
                StaticFactory(Subject)
            ).from_bytes(source[2 * HashPoint.HASH_LENGTH:3 * HashPoint.HASH_LENGTH], resolver),
            ResolverOrigin(Integer.factory(), source[3 * HashPoint.HASH_LENGTH:], resolver).hash_point()
        )

    def without_miner(self) -> 'BankState':
        return BankState(
            self.minted,
            self.used,
            NullableReference(Null(), self.miner.factory),
            self.length
        )

    async def advance(self) -> 'BankState':
        return BankState(
            self.minted,
            self.used,
            NullableReference(Null(), self.miner.factory),
            HashPoint.of(Integer((await self.length.resolve()).integer + 1))
        )

    def minted_tree(self) -> ActiveBinaryTree[Coin, Integer]:
        return ActiveBinaryTree(
            AVL(HashComparator(Fail())), self.minted
        )

    def used_tree(self) -> ActiveBinaryTree[Coin, Integer]:
        return ActiveBinaryTree(
            AVL(HashComparator(Fail())), self.used
        )

    async def use_coins(self, coins: AsyncIterable[HashPoint[Coin]]) -> 'BankState':
        minted: ActiveBinaryTree[Coin, Integer] = self.minted_tree()
        used: ActiveBinaryTree[Coin, Integer] = self.used_tree()
        async for in_coin in coins:
            minted_contains, used_contains, used = await gather(
                minted.contains(in_coin),
                used.contains(in_coin),
                used.add(in_coin)
            )
            assert_true(minted_contains)
            assert_false(used_contains)
            assert isinstance(used, ActiveBinaryTree)
        return BankState(self.minted, used.reference, self.miner, self.length)

    async def _mint_coins(
            self,
            transaction: Transaction
    ) -> 'BankState':
        assert isinstance(transaction, Transaction)
        miner = self.miner_nullable()
        assert isinstance(miner, Nullable)
        minted: ActiveBinaryTree[Coin, Integer] = self.minted_tree()
        assert isinstance(minted, ActiveBinaryTree)
        async for coin, miner in transaction.iter_coins(self.mint(), miner):
            minted_contains, minted = await gather(
                minted.contains(HashPoint.of(coin)),
                minted.add(HashPoint.of(coin)),
            )
            assert_false(minted_contains)
            assert isinstance(minted, ActiveBinaryTree)
        return BankState(minted.reference, self.used, NullableReference(miner, self.miner.factory), self.length)

    async def mint_coins(
            self,
            transaction: Transaction
    ) -> 'BankState':
        assert isinstance(transaction, Transaction)
        verified, bank_state = await gather(
            self.verify(transaction),
            self._mint_coins(transaction),
        )
        assert_true(verified)
        assert isinstance(bank_state, BankState)
        return bank_state

    def miner_nullable(self) -> Nullable[HashPoint[Subject]]:
        return self.miner.reference

    def mint(self) -> int:
        if self.miner_nullable().null():
            return MINT_CONST
        else:
            return 0

    async def verify(self, transaction: Transaction) -> bool:
        assert isinstance(transaction, Transaction)
        assert_true(await transaction.verify(self.mint()))
        return True

    async def _push(self, transaction: Transaction) -> 'BankState':
        assert isinstance(transaction, Transaction)
        return await (
            await self.use_coins(
                (await transaction.data_resolved()).iter_in_coins()
            )
        ).mint_coins(
            transaction
        )

    async def push(self, transaction: HashPoint[Transaction]) -> 'BankState':
        return await self._push(await transaction.resolve())

    async def str(self, tab: int) -> str:
        assert isinstance(tab, int)
        miner_str, minted_str, used_str, length_str = await gather(
            self.miner.str(tab + 1),
            self.minted.str(tab + 1),
            self.used.str(tab + 1),
            hash_point_format(self.length, tab + 1),
        )
        assert isinstance(miner_str, str)
        assert isinstance(minted_str, str)
        assert isinstance(used_str, str)
        assert isinstance(length_str, str)
        return f'(' \
               f'{tabulate(tab + 1)}bank' \
               f'{tabulate(tab + 1)}(miner)' \
               f'{tabulate(tab + 1)}{miner_str}' \
               f'{tabulate(tab + 1)}(minted)' \
               f'{tabulate(tab + 1)}{minted_str}' \
               f'{tabulate(tab + 1)}(used)' \
               f'{tabulate(tab + 1)}{used_str}' \
               f'{tabulate(tab + 1)}(length)' \
               f'{tabulate(tab + 1)}{length_str}' \
               f'{tabulate(tab)})'
